
/*
*
* AngularDistance
*
* 2023 (C) Enzo De Bernardini
* PixInsight Resources - pixinsight.com.ar
*
* THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
* AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT THE AUTHOR
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, BUSINESS
* INTERRUPTION; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; AND LOSS OF USE,
* DATA OR PROFITS) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*
* Changelog:
* 1.0.0  - 2021.04.16 - First release.
* 1.1.0  - 2021.04.27 - AutoStretch. Results treebox. Store, remove, export.
* 1.1.1  - 2023.11.13 - Fix: cursor style after move.
*
*/

#include <pjsr/Sizer.jsh>
#include <pjsr/NumericControl.jsh>
#include <pjsr/FrameStyle.jsh>
#include <pjsr/TextAlign.jsh>
#include <pjsr/StdCursor.jsh>
#include <pjsr/UndoFlag.jsh>
#include <pjsr/StarDetector.jsh>
#include <pjsr/FontFamily.jsh>
#include <pjsr/SampleType.jsh>
#include <pjsr/DataType.jsh>

#include "AutoStretch.js" // Copy from AutoStretch.js script by Andres del Pozo (/scripts/AdP/)

#feature-id       Image Analysis > AngularDistance
#feature-info     <p><b>AngularDistance</b></p><p>Measures the angular distance and position angle between two points in a solved image.</p><p>2021 - Enzo De Bernardini</p><p>pixinsight.com.ar</p>
#feature-icon     AngularDistance.svg

#define TITLE     "AngularDistance"
#define VERSION   "1.1.1"


// Point Selection Modes (array)
var selectionModes = [
    ['Star Detection', 1],
    ['Arbitrary Point', 2]
];

// Alpha (R.A.) Format
var formatModesAlpha = [
    ['Decimal Degrees', 1],
    ['HHh MMm SSs', 2]
];

// Delta (Dec) Format
var formatModesDelta = [
    ['Decimal Degrees', 1],
    ['DDº MM\' SS"', 2]
];

// Separation Format
var formatModesSep = [
    ['Decimal Degrees', 1],
	['Decimal Minutes', 2],
	['Decimal Seconds', 3],
    ['DDº MM\' SS"', 4]
];

var scriptParameters = {
	debug              : false,
	imageID            : "",
	imageActiveWindow  : "",
	zoomFactor         : 1,
	pointSelectionMode : selectionModes[0][1],
	formatAlpha        : formatModesAlpha[1][1],
	formatDelta        : formatModesDelta[1][1],
	formatSeparation   : formatModesSep[3][1],
	pointSelectionBox  : 30, // px
	pointsCelestial    : new Array(),
	outputDirectory    : ""
}

var layoutFlag     = 1;
var outputData     = [];
var lastResultData = "";

//
// DIALOG
//

function mainDialog() {

	// -----------------------------------------------------------------------------------------
	// Adopted pattern to inject all Dialog properties
	this.__base__ = Dialog;
	this.__base__();
	// -----------------------------------------------------------------------------------------

	// Resizable by user? (only set false if scaledMinWidth has a value)
	this.userResizable = false;

	// dialog min width
	this.scaledMinWidth  = 1400;
	this.scaledMinHeight = 900;

	// Window Title
	this.windowTitle = TITLE + " Script";

	// Title
	this.titleLabel              = new Label(this);
	this.titleLabel.frameStyle   = FrameStyle_Sunken;
	this.titleLabel.margin       = 10;
	this.titleLabel.wordWrapping = true;
	this.titleLabel.useRichText  = true;
	this.titleLabel.text = "<p><b>" + TITLE + " " + VERSION + "</b> &mdash; " + "Angular separation and position angle between two points. CTRL + click to select points. Mouse middle button to pan.</p>";


	//
	// Scroll Preview
	//

	this.ScrollControl = new ScrollBox;

	this.ScrollControl.bmp = scriptParameters.imageImage;

	//this.ScrollControl.bmp = new Bitmap( scriptParameters.imageImage.width, scriptParameters.imageImage.height );
	//this.ScrollControl.bmp.assign( scriptParameters.imageImage );

	//this.ScrollControl.bmp.selectedPoint = new Point(scriptParameters.imageImage.width/2, scriptParameters.imageImage.height/2);
	//this.ScrollControl.bmp.selectedPoint = new Point(0, 0);

	//scriptParameters.imageMainView.beginProcess(UndoFlag_NoSwapFile);
	//this.ScrollControl.bmp.assign(scriptParameters.imageImage);
	//this.ScrollControl.bmp.apply(scriptParameters.imageImage);
	//scriptParameters.imageMainView.endProcess();


	this.ScrollControl.autoScroll = true;
	this.ScrollControl.tracking = true;
	this.ScrollControl.setVariableWidth();
	this.ScrollControl.setMinSize(1100, 700);

	var dragging = false;
	this.ScrollControl.dragOrigin = new Point(0);

	this.ScrollControl.viewport.cursor = new Cursor(StdCursor_Cross);

	this.ScrollControl.initScrollBars = (scrollPoint = null) => {

		//this.pageWidth = this.ScrollControl.bmp.width;
		//this.pageHeight = this.ScrollControl.bmp.height;

		this.ScrollControl.setHorizontalScrollRange(0, Math.max(this.ScrollControl.bmp.width - this.ScrollControl.viewport.width, (this.ScrollControl.bmp.width * scriptParameters.zoomFactor) / scriptParameters.zoomFactor - this.ScrollControl.viewport.width / scriptParameters.zoomFactor));
		this.ScrollControl.setVerticalScrollRange(0, Math.max(this.ScrollControl.bmp.height - this.ScrollControl.viewport.height, (this.ScrollControl.bmp.height * scriptParameters.zoomFactor) / scriptParameters.zoomFactor - this.ScrollControl.viewport.height / scriptParameters.zoomFactor));

		/*
		if(scrollPoint == null) {

			// Image Center
			this.ScrollControl.setScrollPosition((this.ScrollControl.bmp.width / 2) - (this.ScrollControl.viewport.width / 2), (this.ScrollControl.bmp.height / 2) - (this.ScrollControl.viewport.height / 2));

		} else {

			//Console.writeln((this.ScrollControl.bmp.width * scriptParameters.zoomFactor)-(scrollPoint.x*scriptParameters.zoomFactor));
			this.ScrollControl.setScrollPosition((this.ScrollControl.bmp.width * scriptParameters.zoomFactor) - (scrollPoint.x * scriptParameters.zoomFactor), (this.ScrollControl.bmp.height * scriptParameters.zoomFactor) - (scrollPoint.y * scriptParameters.zoomFactor));

		};
		*/

		this.ScrollControl.viewport.update();

	};

	this.ScrollControl.viewport.onResize = () => {
		this.ScrollControl.initScrollBars(this.ScrollControl.scrollPosition);
	};

	this.ScrollControl.onHorizontalScrollPosUpdated = (x) => {

		this.ScrollControl.update();

	};

	this.ScrollControl.onVerticalScrollPosUpdated = (y) => {

		this.ScrollControl.update();

	};

	this.ScrollControl.viewport.onMousePress = (x, y, button, buttons, modifiers) => {

		this.ScrollControl.viewport.cursor = new Cursor( StdCursor_ClosedHand );

		this.ScrollControl.dragOrigin.x = x;
		this.ScrollControl.dragOrigin.y = y;
		dragging = true;

	};

	this.ScrollControl.viewport.onMouseMove = (x, y, buttons, modifiers) => {

		with(this) {

			if(dragging) {

				// this.ScrollControl.scrollPosition : scroll location
				// this.ScrollControl.dragOrigin     : cursor location inside viewport
				// x, y                              : cursor displacement quantity

				//Console.writeln(this.ScrollControl.scrollPosition + " -> " + this.ScrollControl.dragOrigin + " -> " + x + " -> " + Math.round((this.ScrollControl.dragOrigin.x - x) / scriptParameters.zoomFactor));

				this.ScrollControl.scrollPosition = new Point(this.ScrollControl.scrollPosition).translatedBy(this.ScrollControl.dragOrigin.x - x, this.ScrollControl.dragOrigin.y - y);
				//this.ScrollControl.scrollPosition = new Point(this.ScrollControl.scrollPosition.x, this.ScrollControl.scrollPosition.y).translatedBy( Math.round((this.ScrollControl.dragOrigin.x - x) / scriptParameters.zoomFactor), Math.round((this.ScrollControl.dragOrigin.y - y) / scriptParameters.zoomFactor));

				this.ScrollControl.dragOrigin.x = x;
				this.ScrollControl.dragOrigin.y = y;

			} else {

				this.ScrollControl.viewport.cursor = new Cursor( StdCursor_Cross );

				//
				// Cursor Coordinates
				//

				var img_x = (x / scriptParameters.zoomFactor) + this.ScrollControl.scrollPosition.x;
				var img_y = (y / scriptParameters.zoomFactor) + this.ScrollControl.scrollPosition.y;

				var pointSelectedCelestial = getCelestial(img_x, img_y); // x,y

				// Right Ascension Format
				switch(scriptParameters.formatAlpha) {

					// Decimal Degrees
					case 1:

						this.cursorDataLabelAlpha.text = pointSelectedCelestial.x.toFixed(6) + "º";

					break;

					// HH:MM:SS
					case 2:

						var alphaHMS = degreesToHMS(pointSelectedCelestial.x);
						this.cursorDataLabelAlpha.text = zeroPadding(alphaHMS[0]) + "<sup>h</sup> " + zeroPadding(alphaHMS[1]) + "<sup>m</sup> " + zeroPadding(alphaHMS[2].toFixed(3)) + "<sup>s</sup>";

					break;

				}; // switch

				// Declination Format
				switch(scriptParameters.formatDelta) {

					// Decimal Degrees
					case 1:

						this.cursorDataLabelDelta.text = pointSelectedCelestial.y.toFixed(6) + "º";


					break;

					// HH:MM:SS
					case 2:

						var deltaDMS = degreesToDMS(pointSelectedCelestial.y);

						if(deltaDMS[0] == -1) {
							var deltaDMSsign = "-";
						} else {
							var deltaDMSsign = "+";
						}

						this.cursorDataLabelDelta.text = deltaDMSsign + zeroPadding(deltaDMS[1]) + "º " + zeroPadding(deltaDMS[2]) + "' " + zeroPadding(deltaDMS[3].toFixed(3)) + "\"";

					break;

				}; // switch

			}; // if dragging

		};

	};

	this.ScrollControl.viewport.onMouseRelease = (x, y, button, buttons, modifiers) => {

		this.ScrollControl.viewport.cursor = new Cursor( StdCursor_Cross );

		if(modifiers == 2) { // CTRL + click

			var img_x = (x / scriptParameters.zoomFactor) + this.ScrollControl.scrollPosition.x;
			var img_y = (y / scriptParameters.zoomFactor) + this.ScrollControl.scrollPosition.y;

			var pointsCelestialPushFlag = true;

			switch(scriptParameters.pointSelectionMode) {

				// Star Detection
				case 1:

					var closestPointCorrected = starCentroid(img_x, img_y);

					if(closestPointCorrected.length > 0) {

						var pointSelectedCelestial = getCelestial(closestPointCorrected[0].x, closestPointCorrected[0].y); // x,y to celestial
						var closestPointObj = { pos: pointSelectedCelestial, imgpos: new Point(img_x, img_y), size: closestPointCorrected[1], box: closestPointCorrected[2] }

					} else {

						pointsCelestialPushFlag = false

					};

				break;

				// Arbitrary
				case 2:

					var pointSelectedCelestial = getCelestial(img_x, img_y); // x,y to celestial
					var closestPointObj = { pos: pointSelectedCelestial, imgpos: new Point(img_x, img_y), size: 100, box: null };

				break;

			}; // switch pointSelectionMode


			// Add new point first to array of points
			if(pointsCelestialPushFlag) {
				scriptParameters.pointsCelestial.unshift(closestPointObj);
			};


			//
			// Distance and PA
			//

			if(scriptParameters.pointsCelestial.length >= 2) {

				var pointFrom = scriptParameters.pointsCelestial[1].pos;
				var pointTo   = scriptParameters.pointsCelestial[0].pos;

				// Distance
				var lastPointsDistance = getDistance(pointFrom, pointTo);

				// Position Angle (PA)
				var lastPointsPA = getPositionAngle(pointFrom, pointTo, lastPointsDistance);

				// Last Result
				lastResultData = {
					ra1:  scriptParameters.pointsCelestial[1].pos.x,
					dec1: scriptParameters.pointsCelestial[1].pos.y,
					ra2:  scriptParameters.pointsCelestial[0].pos.x,
					dec2: scriptParameters.pointsCelestial[0].pos.y,
					x1:   scriptParameters.pointsCelestial[1].imgpos.x,
					y1:   scriptParameters.pointsCelestial[1].imgpos.y,
					x2:   scriptParameters.pointsCelestial[0].imgpos.x,
					y2:   scriptParameters.pointsCelestial[0].imgpos.y,
					sep:  lastPointsDistance,
					pa:   lastPointsPA
				};


				//
				// Label Updates
				//

				// Right Ascension Format
				switch(scriptParameters.formatAlpha) {

					// Decimal Degrees
					case 1:

						this.firstPointAlphaLabel.text  = pointFrom.x.toFixed(6) + "º";
						this.secondPointAlphaLabel.text = pointTo.x.toFixed(6) + "º";

					break;

					// HH:MM:SS
					case 2:

						var pointFromRA = degreesToHMS(pointFrom.x);
						var pointToRA   = degreesToHMS(pointTo.x);

						this.firstPointAlphaLabel.text  = zeroPadding(pointFromRA[0]) + "<sup>h</sup> " + zeroPadding(pointFromRA[1]) + "<sup>m</sup> " + zeroPadding(pointFromRA[2].toFixed(3)) + "<sup>s</sup>";
						this.secondPointAlphaLabel.text = zeroPadding(pointToRA[0]) + "<sup>h</sup> " + zeroPadding(pointToRA[1]) + "<sup>m</sup> " + zeroPadding(pointToRA[2].toFixed(3)) + "<sup>s</sup>";

					break;

				}; // switch


				// Declination Format
				switch(scriptParameters.formatDelta) {

					// Decimal Degrees
					case 1:

						this.firstPointDeltaLabel.text  = pointFrom.y.toFixed(6) + "º";
						this.secondPointDeltaLabel.text = pointFrom.y.toFixed(6) + "º";

					break;

					// DMS
					case 2:

						var pointFromDec = degreesToDMS(pointFrom.y);
						var pointToDec   = degreesToDMS(pointTo.y);

						if(pointFromDec[0] == -1) {
							var pointFromDecSign = "-";
						} else {
							var pointFromDecSign = "+";
						}

						if(pointToDec[0] == -1) {
							var pointToDecSign = "-";
						} else {
							var pointToDecSign = "+";
						}

						this.firstPointDeltaLabel.text  = pointFromDecSign + zeroPadding(pointFromDec[1]) + "º " + zeroPadding(pointFromDec[2]) + "' " + zeroPadding(pointFromDec[3].toFixed(3)) + "\"";
						this.secondPointDeltaLabel.text = pointToDecSign + zeroPadding(pointToDec[1]) + "º " + zeroPadding(pointToDec[2]) + "' " + zeroPadding(pointToDec[3].toFixed(3)) + "\"";

					break;

				}; // switch

				// Separation
				var lastPointsDistanceDMS = degreesToDMS(lastPointsDistance / 60); // lastPointsDistance are in minutes
				this.separationResultLabel.text = "<b>" + lastPointsDistanceDMS[1] + "º " + zeroPadding(lastPointsDistanceDMS[2]) + "' " + lastPointsDistanceDMS[3].toFixed(3) + "\"</b>";

				// Position Angle
				this.paResultLabel.text = "<b>" + lastPointsPA.toFixed(2) + "º</b>";

			};

		};

		this.ScrollControl.viewport.update();

		this.ScrollControl.viewport.cursor = new Cursor( StdCursor_OpenHand );

		dragging = false;

	};


	//
	// MouseWheel Zoom
	//

	this.ScrollControl.viewport.onMouseWheel = (x, y, delta, buttons, modifiers) => {

		if(delta > 0) {
			scriptParameters.zoomFactor++;
		};

		if(delta < 0) {
			scriptParameters.zoomFactor--;
		};

		if(scriptParameters.zoomFactor < 1) {
			scriptParameters.zoomFactor = 1;
		};

		//var point = new Point(this.ScrollControl.scrollPosition.x, this.ScrollControl.scrollPosition.y);
		//var point = new Point((this.ScrollControl.scrollPosition.x - (this.ScrollControl.viewport.width / 2)) + this.ScrollControl.viewport.width, (this.ScrollControl.scrollPosition.y - (this.ScrollControl.viewport.height / 2)) + this.ScrollControl.viewport.height);
		//var point = new Point((this.ScrollControl.scrollPosition.x - (this.ScrollControl.viewport.width / 2)), (this.ScrollControl.scrollPosition.y - (this.ScrollControl.viewport.height / 2)) + this.ScrollControl.viewport.height);

		this.ScrollControl.initScrollBars(new Point(x, y));
		this.ScrollControl.viewport.update();

	};


	this.ScrollControl.viewport.onPaint = (x0, y0, x1, y1) => {


		var g = new VectorGraphics(this.ScrollControl.viewport);

		g.antialiasing        = true;
		g.smoothInterpolation = false;
		g.scaleTransformation(scriptParameters.zoomFactor) // ZOOM

		this.ScrollControl.bmp.selectedRect = new Rect(x0, y0, x1, y1).translated(this.ScrollControl.scrollPosition);

		g.drawBitmap(x0, y0, this.ScrollControl.bmp.render());

		// Text Settings
		var font = new Font(FontFamily_SansSerif, 12, "pt");
		font.bold = true;
		g.font = font;

		if(scriptParameters.pointsCelestial.length > 0) {

			if(scriptParameters.pointsCelestial.length == 1) {
				// Only one point
				var pointsToDraw = [ { pos: scriptParameters.pointsCelestial[0].pos, size: scriptParameters.pointsCelestial[0].size, box: scriptParameters.pointsCelestial[0].box } ];
			} else {
				// Draw latest two points
				var pointsToDraw = [ { pos: scriptParameters.pointsCelestial[1].pos, size: scriptParameters.pointsCelestial[1].size, box: scriptParameters.pointsCelestial[1].box }, { pos: scriptParameters.pointsCelestial[0].pos, size: scriptParameters.pointsCelestial[0].size, box: scriptParameters.pointsCelestial[0].box } ];
			}

			for(var p = 0; p < pointsToDraw.length; p++) {

				switch(p) {
					case 0:  var circleColor = 0xFF954C7F; break;
					case 1:  var circleColor = 0xFF4C9563; break;
					default: var circleColor = 0xFF00FF11; break;
				};

				if(scriptParameters.zoomFactor <= 1) {
					var circleWidth = 2;
				} else {
					var circleWidth = 1;
				}

				var pointImage   = scriptParameters.imageActiveWindow.celestialToImage(pointsToDraw[p].pos.x, pointsToDraw[p].pos.y)
				var circleCenter = new Point(pointImage.x - this.ScrollControl.scrollPosition.x, pointImage.y - this.ScrollControl.scrollPosition.y);
				var circleRadius = Math.sqrt(pointsToDraw[p].size);

				// Star Detection Boxsize (for star detection mode only)
				if(pointsToDraw[p].box != null) {
					g.pen = new Pen(0x40666666, 1);
					g.drawRect((circleCenter.x - (pointsToDraw[p].box / 2)), (circleCenter.y - (pointsToDraw[p].box / 2)), (circleCenter.x + (pointsToDraw[p].box / 2)), (circleCenter.y + (pointsToDraw[p].box / 2)));
				};


				g.pen = new Pen(circleColor, circleWidth);

				// Circle Around Point
				g.drawCircle(circleCenter, circleRadius);

				// Point Pixel
				g.drawPoint(circleCenter);


				// PointText Label
				g.drawText(new Point(pointImage.x - this.ScrollControl.scrollPosition.x + circleRadius + 4, pointImage.y - this.ScrollControl.scrollPosition.y + 6), (p+1).toString());

				// Only latest two points
				if(p == 1) {
					break;
				};

			};

		};

		g.end();

	};

	this.ScrollControl.initScrollBars();



	//
	// FIRST POINT
	//

	// First Point Label
	this.firstPointLabel                      = new Label(this);
	this.firstPointLabel.text                 = "First Point";
	this.firstPointLabel.textAlignment        = TextAlign_Left|TextAlign_VertCenter;

	// First Point Alpha Label
	this.firstPointAlphaLabel                 = new Label(this);
	this.firstPointAlphaLabel.textAlignment   = TextAlign_Right|TextAlign_VertCenter;
	this.firstPointAlphaLabel.textColor       = 0x954C7F;
	this.firstPointAlphaLabel.useRichText     = true;
	this.firstPointAlphaLabel.text            = "-";

	// First Point Delta Label
	this.firstPointDeltaLabel                 = new Label(this);
	this.firstPointDeltaLabel.textAlignment   = TextAlign_Right|TextAlign_VertCenter;
	this.firstPointDeltaLabel.textColor       = 0x954C7F;
	this.firstPointDeltaLabel.text            = "-";

	// First Point Alpha-Delta Vertical
	this.firstPointDataVertical               = new VerticalSizer;
	this.firstPointDataVertical.margin        = 0;
	this.firstPointDataVertical.spacing       = 5;
	this.firstPointDataVertical.add(this.firstPointAlphaLabel);
	this.firstPointDataVertical.add(this.firstPointDeltaLabel);

	// First Point Label + Alpha-Delta Horizontal
	this.firstPointDataHorizontal             = new HorizontalSizer;
	this.firstPointDataHorizontal.margin      = 0;
	this.firstPointDataHorizontal.spacing     = 10;
	this.firstPointDataHorizontal.add(this.firstPointLabel);
	this.firstPointDataHorizontal.add(this.firstPointDataVertical);


	//
	// SECOND POINT
	//

	// Second Point Label
	this.secondPointLabel                     = new Label(this);
	this.secondPointLabel.text                = "Second Point";
	this.secondPointLabel.textAlignment       = TextAlign_Left|TextAlign_VertCenter;

	// Second Point Alpha Label
	this.secondPointAlphaLabel                = new Label(this);
	this.secondPointAlphaLabel.textAlignment  = TextAlign_Right|TextAlign_VertCenter;
	this.secondPointAlphaLabel.textColor      = 0x4C9563;
	this.secondPointAlphaLabel.useRichText    = true;
	this.secondPointAlphaLabel.text           = "-";

	// Second Point Delta Label
	this.secondPointDeltaLabel                = new Label(this);
	this.secondPointDeltaLabel.textAlignment  = TextAlign_Right|TextAlign_VertCenter;
	this.secondPointDeltaLabel.textColor      = 0x4C9563;
	this.secondPointDeltaLabel.text           = "-";

	// Second Point Alpha-Delta Vertical
	this.secondPointDataVertical              = new VerticalSizer;
	this.secondPointDataVertical.margin       = 0;
	this.secondPointDataVertical.spacing      = 5;
	this.secondPointDataVertical.add(this.secondPointAlphaLabel);
	this.secondPointDataVertical.add(this.secondPointDeltaLabel);

	// Second Point Label + Alpha-Delta Horizontal
	this.secondPointDataHorizontal            = new HorizontalSizer;
	this.secondPointDataHorizontal.margin     = 0;
	this.secondPointDataHorizontal.spacing    = 10;
	this.secondPointDataHorizontal.add(this.secondPointLabel);
	this.secondPointDataHorizontal.add(this.secondPointDataVertical);


	//
	// RESULTS
	//

	// Separation Label
	this.separationLabel               = new Label(this);
	this.separationLabel.text          = "Separation";
	this.separationLabel.textAlignment = TextAlign_Left|TextAlign_VertCenter;

	// Separation Result Label
	this.separationResultLabel               = new Label(this);
	this.separationResultLabel.textColor     = 0x3272CD;
	this.separationResultLabel.useRichText   = true;
	this.separationResultLabel.text          = "-";
	this.separationResultLabel.textAlignment = TextAlign_Right|TextAlign_VertCenter;

	// Separation Label + Separation Result
	this.separationDataHorizontal            = new HorizontalSizer;
	this.separationDataHorizontal.margin     = 0;
	this.separationDataHorizontal.spacing    = 10;
	this.separationDataHorizontal.add(this.separationLabel);
	this.separationDataHorizontal.add(this.separationResultLabel);


	// Position Angle Label
	this.paLabel               = new Label(this);
	this.paLabel.text          = "Position Angle (PA)";
	this.paLabel.textAlignment = TextAlign_Left|TextAlign_VertCenter;

	// Position Angle Result Label
	this.paResultLabel               = new Label(this);
	this.paResultLabel.textColor     = 0x3272CD;
	this.paResultLabel.useRichText   = true;
	this.paResultLabel.text          = "-";
	this.paResultLabel.textAlignment = TextAlign_Right|TextAlign_VertCenter;

	// Position Angle Label + Separation Result
	this.paDataHorizontal            = new HorizontalSizer;
	this.paDataHorizontal.margin     = 0;
	this.paDataHorizontal.spacing    = 10;
	this.paDataHorizontal.add(this.paLabel);
	this.paDataHorizontal.add(this.paResultLabel);


	// Mode
	this.pointSelectionModeLabel               = new Label(this);
	this.pointSelectionModeLabel.text          = "Selection Mode";
	this.pointSelectionModeLabel.textAlignment = TextAlign_Left|TextAlign_VertCenter;

	this.pointSelectionModeCombo = new ComboBox(this);

	this.pointSelectionModeCombo.minWidth  = 150;
	this.pointSelectionModeCombo.toolTip = "<p><b>Star Detection</b>: Execute the <i>StarDetector</i> function to locate the selected star centroid (a starlike point is required).</p><p><b>Arbitrary Point</b>: You can select any arbitrary point within the image (they don't have to be a star)</p>";

	with(this.pointSelectionModeCombo) {

		selectionModes.forEach(type => {

			// Fill ComboBox
			addItem(type[0]);

			// Select Item
			if(scriptParameters.pointSelectionMode == type[1]) {
				this.pointSelectionModeCombo.currentItem = type[1]-1;
			};

		});

		onItemSelected = function(index) {

			if(scriptParameters.debug) {
				console.noteln("Point Selection Mode: " + selectionModes[index][1] + " (" + selectionModes[index][0] + ")");
			}; // if debug

			scriptParameters.pointSelectionMode = selectionModes[index][1];

		};

	}; // with


	// Box Size (Star Detector)
	this.boxSizeStarDetectionLabel               = new Label(this);
	this.boxSizeStarDetectionLabel.text          = "Detection Box (px)";
	this.boxSizeStarDetectionLabel.toolTip       = "Star Detection box size in px. Increse for include larger stars.";
	this.boxSizeStarDetectionLabel.textAlignment = TextAlign_Left|TextAlign_VertCenter;

	this.boxSizeStarDetectionSpinbox = new SpinBox(this);

	with(this.boxSizeStarDetectionSpinbox) {

		minValue = 2;
		maxValue = 500;
		stepSize = 2;
		value    = scriptParameters.pointSelectionBox;

		onValueUpdated = (value) =>	{

			scriptParameters.pointSelectionBox = value;

			if(scriptParameters.debug) {
				console.noteln("Star Detector Boxsize: " + value + "px")
			}; // if debug

		};

	};


	// Alpha Format
	this.resultModeAlphaLabel               = new Label(this);
	this.resultModeAlphaLabel.text          = "RA Format (&alpha;)";
	this.resultModeAlphaLabel.useRichText   = true;
	this.resultModeAlphaLabel.toolTip       = "Right Ascension (RA) format.";
	this.resultModeAlphaLabel.textAlignment = TextAlign_Left|TextAlign_VertCenter;

	this.resultModeAlphaModeCombo = new ComboBox(this);
	this.resultModeAlphaModeCombo.minWidth  = 150;

	with(this.resultModeAlphaModeCombo) {

		formatModesAlpha.forEach(type => {

			// Fill ComboBox
			addItem(type[0]);

			// Select Item
			if(scriptParameters.formatAlpha == type[1]) {
				this.resultModeAlphaModeCombo.currentItem = type[1]-1;
			};

		});

		onItemSelected = function(index) {

			if(scriptParameters.debug) {
				console.noteln("Alpha Format: " + formatModesAlpha[index][1]);
			}; // if debug

			scriptParameters.formatAlpha = formatModesAlpha[index][1];

		};

	}; // with


	// Delta Format
	this.resultModeDeltaLabel               = new Label(this);
	this.resultModeDeltaLabel.text          = "Dec Format (&delta;)";
	this.resultModeDeltaLabel.useRichText   = true;
	this.resultModeDeltaLabel.textAlignment = TextAlign_Left|TextAlign_VertCenter;

	this.resultModeDeltaModeCombo = new ComboBox(this);
	this.resultModeDeltaModeCombo.toolTip   = "Declination (Dec) format.";
	this.resultModeDeltaModeCombo.minWidth  = 150;

	with(this.resultModeDeltaModeCombo) {

		formatModesDelta.forEach(type => {

			// Fill ComboBox
			addItem(type[0]);

			// Select Item
			if(scriptParameters.formatDelta == type[1]) {
				this.resultModeDeltaModeCombo.currentItem = type[1]-1;
			};

		});

		onItemSelected = function(index) {

			if(scriptParameters.debug) {
				console.noteln("Delta Format: " + formatModesDelta[index][1]);
			}; // if debug

			scriptParameters.formatDelta = formatModesDelta[index][1];

		};

	}; // with


	// Separation Format
	/*
	this.resultModeSeparationLabel               = new Label(this);
	this.resultModeSeparationLabel.text          = "Separation format";
	this.resultModeSeparationLabel.useRichText   = true;
	this.resultModeSeparationLabel.toolTip       = "Separation format.";
	this.resultModeSeparationLabel.textAlignment = TextAlign_Left|TextAlign_VertCenter;

	this.resultModeSeparationCombo = new ComboBox(this);
	this.resultModeSeparationCombo.minWidth = 150;

	with(this.resultModeSeparationCombo) {

		formatModesSep.forEach(type => {

			// Fill ComboBox
			addItem(type[0]);

			// Select Item
			if(scriptParameters.formatSeparation == type[1]) {
				this.resultModeSeparationCombo.currentItem = type[1]-1;
			};

		});

		onItemSelected = function(index) {

			if(scriptParameters.debug) {
				console.noteln("Separation Format: " + formatModesSep[index][1]);
			}; // if debug

			scriptParameters.formatSeparation = formatModesSep[index][1];

		};

	}; // with
	*/

	// Point Selection Mode Horizontal Sizer
	this.pointSelectionModeGroupBoxH         = new HorizontalSizer;
	this.pointSelectionModeGroupBoxH.margin  = 0;
	this.pointSelectionModeGroupBoxH.spacing = 10;
	this.pointSelectionModeGroupBoxH.add(this.pointSelectionModeLabel);
	this.pointSelectionModeGroupBoxH.add(this.pointSelectionModeCombo);

	// Star Detector Box Size
	this.boxSizeStarDetectionGroupBoxH         = new HorizontalSizer;
	this.boxSizeStarDetectionGroupBoxH.margin  = 0;
	this.boxSizeStarDetectionGroupBoxH.spacing = 10;
	this.boxSizeStarDetectionGroupBoxH.add(this.boxSizeStarDetectionLabel);
	this.boxSizeStarDetectionGroupBoxH.add(this.boxSizeStarDetectionSpinbox);

	// Alpha Result Format Horizontal Sizer
	this.resultAlphaFormatGroupBoxH         = new HorizontalSizer;
	this.resultAlphaFormatGroupBoxH.margin  = 0;
	this.resultAlphaFormatGroupBoxH.spacing = 10;
	this.resultAlphaFormatGroupBoxH.add(this.resultModeAlphaLabel);
	this.resultAlphaFormatGroupBoxH.add(this.resultModeAlphaModeCombo);

	// Delta Result Format Horizontal Sizer
	this.resultDeltaFormatGroupBoxH         = new HorizontalSizer;
	this.resultDeltaFormatGroupBoxH.margin  = 0;
	this.resultDeltaFormatGroupBoxH.spacing = 10;
	this.resultDeltaFormatGroupBoxH.add(this.resultModeDeltaLabel);
	this.resultDeltaFormatGroupBoxH.add(this.resultModeDeltaModeCombo);

	// Separation Result Format Horizontal Sizer
	/*
	this.resultSeparationFormatGroupBoxH         = new HorizontalSizer;
	this.resultSeparationFormatGroupBoxH.margin  = 0;
	this.resultSeparationFormatGroupBoxH.spacing = 10;
	this.resultSeparationFormatGroupBoxH.add(this.resultModeSeparationLabel);
	this.resultSeparationFormatGroupBoxH.add(this.resultModeSeparationCombo);
	*/

	// Right Sidebar
	this.settingsGroupBox               = new GroupBox(this);
	this.settingsGroupBox.title         = "Settings";
	this.settingsGroupBox.sizer         = new VerticalSizer;
	this.settingsGroupBox.sizer.margin  = 10;
	this.settingsGroupBox.sizer.spacing = 10;
	this.settingsGroupBox.sizer.add(this.pointSelectionModeGroupBoxH);
	this.settingsGroupBox.sizer.spacing = 10;
	this.settingsGroupBox.sizer.add(this.boxSizeStarDetectionGroupBoxH);
	this.settingsGroupBox.sizer.spacing = 10;
	this.settingsGroupBox.sizer.add(this.resultAlphaFormatGroupBoxH);
	this.settingsGroupBox.sizer.spacing = 10;
	this.settingsGroupBox.sizer.add(this.resultDeltaFormatGroupBoxH);
	//this.settingsGroupBox.sizer.spacing = 10;
	//this.settingsGroupBox.sizer.add(this.resultSeparationFormatGroupBoxH);


	// GROUPBOX
	// Right Sidebar
	this.resultsGroupBox               = new GroupBox(this);
	this.resultsGroupBox.title         = "Results";
	this.resultsGroupBox.sizer         = new VerticalSizer;
	this.resultsGroupBox.sizer.margin  = 10;
	this.resultsGroupBox.sizer.spacing = 10;
	this.resultsGroupBox.sizer.add(this.firstPointDataHorizontal);
	this.resultsGroupBox.sizer.spacing = 20;
	this.resultsGroupBox.sizer.add(this.secondPointDataHorizontal);
	this.resultsGroupBox.sizer.spacing = 20;
	this.resultsGroupBox.sizer.add(this.separationDataHorizontal);
	this.resultsGroupBox.sizer.spacing = 20;
	this.resultsGroupBox.sizer.add(this.paDataHorizontal);


	// Cursor Data RA Label
	this.cursorDataLabelTitleAlpha               = new Label(this);
	this.cursorDataLabelTitleAlpha.textAlignment = TextAlign_Center|TextAlign_VertCenter;
	this.cursorDataLabelTitleAlpha.wordWrapping  = false;
	this.cursorDataLabelTitleAlpha.useRichText   = true;
	this.cursorDataLabelTitleAlpha.text          = "&alpha;";

	// Cursor Data RA
	this.cursorDataLabelAlpha               = new Label(this);
	this.cursorDataLabelAlpha.textAlignment = TextAlign_Center|TextAlign_VertCenter;
	this.cursorDataLabelAlpha.wordWrapping  = false;
	this.cursorDataLabelAlpha.useRichText   = true;
	this.cursorDataLabelAlpha.text          = "-";

	// Cursor Data Dec Label
	this.cursorDataLabelTitleDelta               = new Label(this);
	this.cursorDataLabelTitleDelta.textAlignment = TextAlign_Center|TextAlign_VertCenter;
	this.cursorDataLabelTitleDelta.wordWrapping  = false;
	this.cursorDataLabelTitleDelta.useRichText   = true;
	this.cursorDataLabelTitleDelta.text          = "&delta;";

	// Cursor Data Dec
	this.cursorDataLabelDelta               = new Label(this);
	this.cursorDataLabelDelta.textAlignment = TextAlign_Center|TextAlign_VertCenter;
	this.cursorDataLabelDelta.wordWrapping  = false;
	this.cursorDataLabelDelta.useRichText   = true;
	this.cursorDataLabelDelta.text          = "-";


	// GROUPBOX
	// Cursor Data
	this.cursorDataHorizontal         = new HorizontalSizer;
	this.cursorDataHorizontal.margin  = 10;
	this.cursorDataHorizontal.spacing = 0;
	this.cursorDataHorizontal.add(this.cursorDataLabelTitleAlpha);
	this.cursorDataHorizontal.add(this.cursorDataLabelAlpha);
	this.cursorDataHorizontal.addSpacing(10);
	this.cursorDataHorizontal.add(this.cursorDataLabelTitleDelta);
	this.cursorDataHorizontal.add(this.cursorDataLabelDelta);


	// Measurements Store TreeBox
	this.ResultsTreebox                   = new TreeBox(this);
	this.ResultsTreebox.alternateRowColor = true;
	this.ResultsTreebox.headerVisible     = true;
	this.ResultsTreebox.headerSorting     = false;
	this.ResultsTreebox.indentSize        = 0;
	this.ResultsTreebox.font              = new Font(FontFamily_Monospace, 10);

	this.ResultsTreebox.setColumnWidth(0, 50);

	this.ResultsTreebox.setHeaderText(0,  "#");
	this.ResultsTreebox.setHeaderText(1,  "RA_1");
	this.ResultsTreebox.setHeaderText(2,  "Dec_1");
	this.ResultsTreebox.setHeaderText(3,  "RA_2");
	this.ResultsTreebox.setHeaderText(4,  "Dec_2");
	this.ResultsTreebox.setHeaderText(5,  "x1_px");
	this.ResultsTreebox.setHeaderText(6,  "y1_px");
	this.ResultsTreebox.setHeaderText(7,  "x2_px");
	this.ResultsTreebox.setHeaderText(8,  "y2_px");
	this.ResultsTreebox.setHeaderText(9,  "Sep_min");
	this.ResultsTreebox.setHeaderText(10, "PA");

	this.ResultsTreebox.setHeaderAlignment(0, 1);
	this.ResultsTreebox.setHeaderAlignment(1, 1);
	this.ResultsTreebox.setHeaderAlignment(2, 1);
	this.ResultsTreebox.setHeaderAlignment(3, 1);
	this.ResultsTreebox.setHeaderAlignment(4, 1);
	this.ResultsTreebox.setHeaderAlignment(5, 1);
	this.ResultsTreebox.setHeaderAlignment(6, 1);
	this.ResultsTreebox.setHeaderAlignment(7, 1);
	this.ResultsTreebox.setHeaderAlignment(8, 1);
	this.ResultsTreebox.setHeaderAlignment(9, 1);
	this.ResultsTreebox.setHeaderAlignment(10, 1);


	/*
	this.ResultsTreebox.onNodeClicked = (item, column) => {

		for(var i = 0; i <= this.ResultsTreebox.numberOfChildren - 1; i++) {

			if(this.ResultsTreebox.child(i).selected) {
				Console.writeln(this.ResultsTreebox.child(i).text(column));
				// outputData
			};

		}; // for

	};
	*/


	// Results Store TreeBox Node Creator
	this.addResultsTreeboxNode = (data) => {

		var node       = new TreeBoxNode(this.ResultsTreebox);
		node.checkable = true;
		node.checked   = true;

		node.setText(0,  "");
		node.setText(1,  format("%.5f", data.ra1));
		node.setText(2,  format("%.5f", data.dec1));
		node.setText(3,  format("%.5f", data.ra2));
		node.setText(4,  format("%.5f", data.dec2));
		node.setText(5,  format("%.2f", data.x1));
		node.setText(6,  format("%.2f", data.y1));
		node.setText(7,  format("%.2f", data.x2));
		node.setText(8,  format("%.2f", data.y2));
		node.setText(9,  format("%.3f", data.sep));
		node.setText(10, format("%.2f", data.pa));

		node.setAlignment(0,  1);
		node.setAlignment(1,  2);
		node.setAlignment(2,  2);
		node.setAlignment(3,  2);
		node.setAlignment(4,  2);
		node.setAlignment(5,  2);
		node.setAlignment(6,  2);
		node.setAlignment(7,  2);
		node.setAlignment(8,  2);
		node.setAlignment(9,  2);
		node.setAlignment(10, 2);

		//node.setIcon(0, ":/bullets/bullet-ball-glass-grey.png");

		return node;

	}; // add node


	// Remove Unchecked
	this.removeResultsTreeboxNode = () => {

		for(var i = 0; i <= this.ResultsTreebox.numberOfChildren - 1; i++) {

			if(!this.ResultsTreebox.child(i).checked) {
				this.ResultsTreebox.remove(i);
				i--;
			};

		}; // for

	}; // remove node


	// Button Store
	this.storeResultButton = new PushButton(this);
	this.storeResultButton.text = " Store Result ";
	this.storeResultButton.icon = ":/icons/grid-add.png";
	this.storeResultButton.onClick = () => {

		if(lastResultData != "") {

			this.addResultsTreeboxNode(lastResultData);

		} else {

			var msg = new MessageBox("There are no results to save.", "Export Error", StdIcon_Error, StdButton_Ok);
			msg.execute();

		};

	};


	// Button Remove Unchecked
	this.removeResultButton = new PushButton(this);
	this.removeResultButton.text = " Remove Unchecked ";
	this.removeResultButton.icon = ":/icons/grid-remove.png";
	this.removeResultButton.onClick = () => {

		this.removeResultsTreeboxNode();

	};


	// Export CSV
	this.exportResultButton = new PushButton(this);
	this.exportResultButton.text = " Export Results to CSV ";
	this.exportResultButton.icon = ":/icons/document-csv.png";
	this.exportResultButton.onClick = () => {

		if(this.ResultsTreebox.numberOfChildren > 0) {

			if(scriptParameters.outputDirectory != "") {

				var l          = 0;
				var data       = new Array();
				var outputData = new Array();

				//
				// Columns Names
				//

				for(var c = 1; c <= this.ResultsTreebox.numberOfColumns - 1; c++) {
					data[c-1] = this.ResultsTreebox.headerText(c);
				};

				outputData[0] = data.toString();
				l++;


				//
				// Node Data
				//

				for(var n = 0; n <= this.ResultsTreebox.numberOfChildren - 1; n++) {

					if(this.ResultsTreebox.child(n).checked) {

						for(var c = 1; c <= this.ResultsTreebox.numberOfColumns - 1; c++) {
							data[c-1] = this.ResultsTreebox.child(n).text(c);
						};

						outputData[l] = data.toString();

						l++;

					};

				}; // for


				//
				// CSV Save
				//

				var csv = new saveCSV(scriptParameters.outputDirectory, outputData);
				csv.write();

				if(scriptParameters.debug) {
					Console.writeln("Output Data:");
					Console.writeln(JSON.stringify(outputData, null, 4));
				}; //if debug

			} else {

				var msg = new MessageBox("Please select a destination directory to save the file.", "Export Error", StdIcon_Error, StdButton_Ok);
				msg.execute();

			}; // if

		} else {

			var msg = new MessageBox("There are no results to export.", "Export Error", StdIcon_Error, StdButton_Ok);
			msg.execute();

		}; // if

	};


	// Edit
	// Output Directory
	this.outputDirEdit          = new Edit(this);
	this.outputDirEdit.readOnly = true;
	this.outputDirEdit.text     = "";


	// Label + Button
	// Select Output Directory Label + Button

	this.outputExtLabel                = new Label(this);
	this.outputExtLabel.text           = "Output Directory";
	this.outputExtLabel.textAlignment  = TextAlign_Left|TextAlign_VertCenter;

	this.outputDirSelectButton = new ToolButton( this );
	this.outputDirSelectButton.icon = this.scaledResource(":/icons/folder.png");
	this.outputDirSelectButton.setScaledFixedSize(20, 20);
	this.outputDirSelectButton.toolTip = "<p>Select the output directory.</p>";

	this.outputDirSelectButton.onClick = function() {

		var gdd = new GetDirectoryDialog;
		gdd.caption = "CSV output directory";

		if(gdd.execute()) {
			scriptParameters.outputDirectory = gdd.directory;
			this.dialog.outputDirEdit.text  = gdd.directory;
		}

	};


	// HORIZONTAL -- Output Directory
	this.outputDirectoryGroupBoxH         = new HorizontalSizer;
	this.outputDirectoryGroupBoxH.margin  = 0;
	this.outputDirectoryGroupBoxH.spacing = 10;
	this.outputDirectoryGroupBoxH.add(this.outputDirEdit);
	this.outputDirectoryGroupBoxH.add(this.outputDirSelectButton);


	// GROUPBOX
	// Export
	this.exportGroupBox               = new GroupBox(this);
	this.exportGroupBox.title         = "Export";
	this.exportGroupBox.sizer         = new VerticalSizer;
	this.exportGroupBox.sizer.margin  = 10;
	this.exportGroupBox.sizer.spacing = 10;
	this.exportGroupBox.sizer.add(this.storeResultButton);
	this.exportGroupBox.sizer.spacing = 20;
	this.exportGroupBox.sizer.add(this.removeResultButton);
	this.exportGroupBox.sizer.spacing = 20;
	this.exportGroupBox.sizer.add(this.exportResultButton);
	this.exportGroupBox.sizer.spacing = 20;
	this.exportGroupBox.sizer.add(this.outputExtLabel);
	this.exportGroupBox.sizer.spacing = 10;
	this.exportGroupBox.sizer.add(this.outputDirectoryGroupBoxH);


	// Button AutoStretch
	this.autoStretchButton = new PushButton(this);
	this.autoStretchButton.text = " Auto Stretch ";
	this.autoStretchButton.icon = ":/icons/burn.png";
	this.autoStretchButton.onClick = () => {

		let autoSTF = new AutoStretch();
		autoSTF.HardApply(scriptParameters.workingPreview.mainView, false);
		this.ScrollControl.viewport.update();

	};


	// Button Layout Toggle
	this.layoutToggleButton = new PushButton(this);
	this.layoutToggleButton.text = " Layout Toggle ";
	this.layoutToggleButton.icon = ":/icons/window-arrange-vert.png";
	this.layoutToggleButton.onClick = () => {

		switch(layoutFlag) {
			case 1: this.ScrollControl.setMinHeight(400); layoutFlag = 2; break;
			case 2: this.ScrollControl.setMinHeight(700); layoutFlag = 1; break;
		}


	};


	// Button Exit
	this.cancelButton = new PushButton(this);
	this.cancelButton.text = "Exit";
	this.cancelButton.icon = ":/toolbar/file-exit.png";
	this.cancelButton.onClick = () => {

		var msg = new MessageBox("Exit?", "Confirm Exit", StdIcon_Question, StdButton_Yes, StdButton_No);

		if(msg.execute() == StdButton_Yes) {
			this.cancel();
		}

	};


	// PixInsight Resources button
	this.PIRButton         = new ToolButton(this);
	this.PIRButton.icon    = this.scaledResource(":/toolbar/resources-pixinsight-resources.png");
	this.PIRButton.toolTip = "Open PixInsight Resources website on PixInsight web browser.";
	this.PIRButton.onMousePress = () => {

		Console.show();
		Console.warningln("\n    ____     ____    ____ \n   / __ \\\   /  _/   / __ \\\  \n  / /_/ /   / /    / /_/ /\n / ____/  _/ /    / _, _/ \n/_/      /___/   /_/ |_|  \n")
		Console.noteln("\nPixInsight Resources: https://pixinsight.com.ar");

		Dialog.openBrowser( "https://pixinsight.com.ar/", "PixInsight Resources", 1200, 800);

	};


	//
	// SIZERS
	//

	// HORIZONTAL -- Buttons
	this.utilButtonSizer = new HorizontalSizer;
	this.utilButtonSizer.margin = 0;
	this.utilButtonSizer.add(this.autoStretchButton);
	this.utilButtonSizer.addSpacing(10);
	this.utilButtonSizer.add(this.layoutToggleButton);

	// HORIZONTAL -- Bottom icons, buttons
	this.execButtonSizer = new HorizontalSizer;
	this.execButtonSizer.margin = 0;
	this.execButtonSizer.add(this.cancelButton);
	this.execButtonSizer.addSpacing(10);
	this.execButtonSizer.add(this.PIRButton);

	// VERTICAL - Right Sidebar
	this.rightSidebarVertical = new VerticalSizer;
	this.rightSidebarVertical.margin = 0;
	this.rightSidebarVertical.add(this.settingsGroupBox);
	this.rightSidebarVertical.addSpacing(10);
	this.rightSidebarVertical.add(this.resultsGroupBox);
	this.rightSidebarVertical.addSpacing(10);
	this.rightSidebarVertical.add(this.exportGroupBox);
	this.rightSidebarVertical.addSpacing(10);
	this.rightSidebarVertical.add(this.utilButtonSizer);
	this.rightSidebarVertical.addStretch();
	this.rightSidebarVertical.add(this.cursorDataHorizontal);
	this.rightSidebarVertical.addSpacing(10);
	this.rightSidebarVertical.add(this.execButtonSizer);

	// VERTICAL - Left Body
	this.leftBodyVertical = new VerticalSizer();
	this.leftBodyVertical.margin = 0;
	this.leftBodyVertical.add(this.ScrollControl);
	this.leftBodyVertical.addSpacing(10);
	this.leftBodyVertical.add(this.ResultsTreebox);

	// HORIZONTAL -- Main Grid
	this.mainHorizontal = new HorizontalSizer;
	this.mainHorizontal.margin = 0;
	this.mainHorizontal.add(this.leftBodyVertical);
	this.mainHorizontal.addSpacing(10);
	this.mainHorizontal.add(this.rightSidebarVertical);

	// VERTICAL
	this.sizer = new VerticalSizer;
	this.sizer.margin = 10;
	this.sizer.add(this.titleLabel);
	this.sizer.addSpacing(10);
	this.sizer.add(this.mainHorizontal);

}; // mainDialog

mainDialog.prototype = new Dialog;


//
// Functions
//

function getCelestial(x, y) {

	return scriptParameters.imageActiveWindow.imageToCelestial(x, y);

};


function getDistance(pointFrom, pointTo) {

	// pointFrom, pointTo = in degrees

	var term_1 = Math.cos(Math.min2rad(5400) - Math.min2rad(60 * pointFrom.y));
	var term_2 = Math.cos(Math.min2rad(5400) - Math.min2rad(60 * pointTo.y));
	var term_3 = Math.sin(Math.min2rad(5400) - Math.min2rad(60 * pointFrom.y));
	var term_4 = Math.sin(Math.min2rad(5400) - Math.min2rad(60 * pointTo.y));
	var term_5 = Math.cos(Math.min2rad(60 * pointFrom.x) - Math.min2rad(60 * pointTo.x));

	var res_1 = (term_1 * term_2) + (term_3 * term_4 * term_5);
	var res_f = Math.rad2min(Math.acos(res_1));

	return res_f; // arc min

};


function getPositionAngle(pointFrom, pointTo, distance) {

	// pointFrom, pointTo = in degrees
	// distance           = in minutes

	// RA diff
	var term_aux_1 = pointTo.x - pointFrom.x;

	// For negative RA
	if(term_aux_1 < 0) {
		term_aux_1 += 360;
	};

	var term_1 = Math.cos(Math.min2rad(60 * pointFrom.y)) * Math.cos(Math.min2rad(60 * pointTo.y)) * Math.sin(Math.min2rad(60 * (term_aux_1)));
	var term_2 = Math.sin(Math.min2rad(60 * pointTo.y)) - (Math.cos(Math.min2rad(distance)) * Math.sin(Math.min2rad(60 * pointFrom.y)));
	var res_f  = (Math.rad2min(Math.atan(term_1 / term_2)) / 60);

	// Quadrant Selection
	if(res_f < 0) {

		if(pointFrom.y > pointTo.y) {
			// SE
			res_f += 180;
		} else {
			// NW
			res_f += 360;
		};

	} else {

		if(pointFrom.x > pointTo.x) {
			// SW
			res_f += 180;
		};

	};

	return res_f;

}


function starCentroid(img_x, img_y) {

	//
	// Preview box around selected point
	//

	var box     = scriptParameters.pointSelectionBox;
	var boxHalf = Math.floor((box / 2));

	// Adjust between preview and source image
	var xDiff = img_x - Math.floor(img_x);
	var yDiff = img_y - Math.floor(img_y);

	var x0 = Math.floor(img_x) - boxHalf;
	var y0 = Math.floor(img_y) - boxHalf;

	var x1 = Math.floor(img_x) + boxHalf;
	var y1 = Math.floor(img_y) + boxHalf;

	scriptParameters.imageActiveWindow.createPreview(x0, y0, x1, y1, "StarPreviewTemp");

	var previews = scriptParameters.imageActiveWindow.previews; // array

	// Image from preview
	/*
	var pointSelectionBox = new ImageWindow(box, box, scriptParameters.imageImage.numberOfChannels, scriptParameters.imageImage.bitsPerSample, scriptParameters.imageImage.sampleType == SampleType_Real, false, "StarImageTemp");

	with(pointSelectionBox.mainView) {
		beginProcess(UndoFlag_NoSwapFile);
		image.assign(previews[previews.length-1].image);
		endProcess();
	}
   */

	var starDetection = new StarDetector();

	var stars = starDetection.stars(previews[previews.length-1].image);

	var closestPointCorrected = new Array();

	var previewView = new View(previews[previews.length-1]);

	if(stars.length > 0) {

		//Console.noteln(JSON.stringify(stars, null, 4));

		// Find the closest detected star to the clicked point (center of preview = half box size)
		var closestPoint = findClosestPair(stars, new Array((box / 2), (box / 2)));

		// Search for the selected star's data (generated by StarDetector)
		for(var s = 0; s < stars.length; s++) {

			if(stars[s].pos.x == closestPoint[0] && stars[s].pos.y == closestPoint[1]) {
				var closestPointSize = stars[s].size;
				break;
			};

		}; // for


		// Point Correction (x,y preview to x,y main image)
		closestPointCorrected.push(new Point((closestPoint[0] + (img_x-(box/2)) - xDiff), (closestPoint[1] + (img_y-(box/2))) - yDiff), closestPointSize, box);

		if(scriptParameters.debug) {
			Console.noteln("Clicked Point x, y           : " + img_x + ", " + img_y);
			Console.noteln("Diff x,y                     : " + xDiff + ", " + yDiff);
			Console.noteln("Preview x0, y0               : " + x0 + "," + y0);
			Console.noteln("Closest Point x, y           : " + closestPoint[0] + ", " + closestPoint[1]);
			Console.noteln("Closest Point Corrected x, y : " + closestPointCorrected[0].x + ", " + closestPointCorrected[0].y);
			Console.noteln("\n");
		};

	} else {

		var msg = new MessageBox("<p>No stars have been detected.</p>", TITLE, StdIcon_Warning, StdButton_Ok);
		msg.execute();

	}; // if stars detected

	// Delete Temporal Preview
	scriptParameters.imageActiveWindow.deletePreview(previewView);

	return closestPointCorrected;

};


function findClosestPair(stars, search_element) {

	// Stars Array x,y position
	var elements = new Array();

	for(var s = 0; s < stars.length; s++) {

		elements[s] = new Array(stars[s].pos.x, stars[s].pos.y);

		if(scriptParameters.debug) {
			Console.noteln("Star " + s + ": " + stars[s].pos.x + ", " + stars[s].pos.y);
		};

	}; // for

    return elements.reduce(function (a, b) {
        function getDelta(v, w) {
            return Math.abs(v[0] - w[0]) * Math.abs(v[1] - w[1]);
        }
        return getDelta(a, search_element) < getDelta(b, search_element) ? a : b;
    });

};


function degreesToHMS(degrees) {

	var hDecimal = (degrees * 24) / 360;  // decimal hours
	var h        = Math.trunc(hDecimal);  // int hours
	var mDecimal = (hDecimal - h) * 60    // decimal minutes
	var m        = Math.trunc(mDecimal);  // int minutes
	var s        = (mDecimal - m) * 60    // decimal seconds

	return new Array(h, m, s);

};


function degreesToDMS(degrees) {

	return Math.decimalToSexagesimal(degrees);

};


function zeroPadding(number) {

	if(number < 10) {
		number = "0" + number.toString();
	};

	return number;

};


// Adapted from https://pixinsight.com/forum/index.php?threads/saving-in-csv.6623/post-46316
function saveCSV(filePath, rows) {

	if(filePath != "") {

		this.filePath = filePath;

		if(rows == undefined ) {
			this.rows = new Array;
		} else {
			this.rows = rows;
		};

		// Writes array to lines
		this.write = function() {

			function outputCSVSequence(file, items, title) {

				//file.outTextLn(format("%s,%d", title, items.length));

				if(items.length > 0) {

					for(var i = 0;;) {

						file.outText(items[i]);

						if(++i == items.length) {
							break;
						}; // if

						file.outText('\n');

					}; // ofr

					file.outTextLn('');

				}; // if items

			}; // function

			// Validation: output directory final slash
			if(this.filePath.charAt(this.filePath.length - 1) != "/") {
				this.filePath += "/";
			}; // if final slash

			var file = new File;
			file.createForWriting(this.filePath + "AngularDistance-data.csv");
			outputCSVSequence(file, this.rows);
			file.close();

		};

	} else {

		Console.writeln("No path");

	}; // if file path

}


/*
function pad(pad, str, padLeft) {
  if (typeof str === 'undefined')
    return pad;
  if (padLeft) {
    return (pad + str).slice(-pad.length);
  } else {
    return (str + pad).substring(0, pad.length);
  }
}
*/


//
// MAIN
//

function main () {

	// ImageWindow
	scriptParameters.imageActiveWindow = ImageWindow.activeWindow;

	// Validations
	if(scriptParameters.imageActiveWindow.isNull) {

		var msg = new MessageBox("<p>No active image.</p>", TITLE, StdIcon_Warning, StdButton_Ok);
		msg.execute();

	} else if(scriptParameters.imageActiveWindow.currentView.isPreview) {

		var msg = new MessageBox("<p>This script can not work on prewiews</p>", TITLE, StdIcon_Warning, StdButton_Ok);
		msg.execute();

	} else {

		// This script need a solved image
		if(scriptParameters.imageActiveWindow.astrometricSolutionSummary() != "") {

			// Debug?
			if(scriptParameters.debug) {
				Console.show();
				Console.noteln("*** DEBUG MODE ON ***");
			} else {
				Console.hide();
			}; // if debug

			scriptParameters.imageMainView = scriptParameters.imageActiveWindow.mainView;
			scriptParameters.imageImage    = scriptParameters.imageActiveWindow.mainView.image;

			scriptParameters.imageView     = new View(scriptParameters.imageMainView);
			scriptParameters.imageID       = scriptParameters.imageView.id;


			// Working Preview to show in preview viewport
			// (only visual, data are getted from main source image)

			scriptParameters.workingPreview = new ImageWindow(
				1, 1, 1, scriptParameters.imageMainView.image.bitsPerSample,
				scriptParameters.imageMainView.image.sampleType == SampleType_Real,
				false, "AngularDistancePreviewImage");

			with(scriptParameters.workingPreview.mainView)
			{
				beginProcess(UndoFlag_NoSwapFile);
				image.assign(scriptParameters.imageMainView.image);
				endProcess();
			}

			if(scriptParameters.debug) {
				scriptParameters.workingPreview.show();
				scriptParameters.workingPreview.zoomToFit();
			};


			scriptParameters.imageMainView = scriptParameters.workingPreview.mainView;
			scriptParameters.imageImage    = scriptParameters.workingPreview.mainView.image;


			// Create and show the dialog
			let dialog = new mainDialog;

			if(dialog.execute()) {

				if(scriptParameters.debug) {
					Console.noteln("Execute -- OK");
				};

			} else {

				if(scriptParameters.debug) {
					Console.noteln("Execute -- CANCEL");
				};


				// Close Preview Image
				scriptParameters.workingPreview.forceClose();


			}; // if execution

		} else {

			var msg = new MessageBox("<p>The image has no astrometric solution.</p>", TITLE, StdIcon_Warning, StdButton_Ok);
			msg.execute();

		};

	};

}; // main


main();


// Resistance is futile
